from enum import Enum

from .eorzea_time import EorzeaTime
from .eorzea_weather import EorzeaWeather


__title__ = 'EorzeaEnv'
__author__ = 'Elton H.Y. Chou'
__license__ = 'MIT'
__version__ = '1.4.4'


class EorzeaLang(str, Enum):
    EN = "en"
    JA = "ja"
    DE = "de"
    FR = "fr"


import pytest
from EorzeaEnv import EorzeaLang, EorzeaTime, EorzeaWeather


class TestForecast:
    def test_forecast(self):
        assert EorzeaWeather.forecast(
            'Eureka Pagos', 1542651599.999
        ) == 'Fog'
        assert EorzeaWeather.forecast(
            'Eureka Pyros', 1542591400.045
        ) == 'Umbral Wind'
        assert EorzeaWeather.forecast(
            'Sigmascape V4.0', 1542591400.045
        ) == 'Dimensional Disruption'
        assert EorzeaWeather.forecast(
            'Bowl of Embers', 1542591400.045
        ) == 'Heat Waves'
        assert EorzeaWeather.forecast(
            'the ruby sea', 1542591400.045
        ) == 'Fair Skies'
        assert EorzeaWeather.forecast(
            'sea of clouds', 1542591400.045
        ) == 'Fog'
        assert EorzeaWeather.forecast(
            'asdfUpper aetheroacoustic exploratory sitea dsf',
            1542591400.045,
            strict=False
        ) == 'Fair Skies'

    def test_field(self):
        weathers = EorzeaWeather.forecast(
            'Eureka Pyros', EorzeaTime.weather_period(10))

        for weather in weathers:
            assert isinstance(weather, str)

        with pytest.raises(ValueError):
            pyros_weather = EorzeaWeather.forecast(
                'EuPyros', 1542591400.045
            )

    def test_localize(self):
        assert EorzeaWeather.forecast(
            'Eureka Pagos',
            1542651599.999,
            lang=EorzeaLang.EN
        ) == 'Fog'
        assert EorzeaWeather.forecast(
            'Eureka Pagos',
            1542651599.999,
            lang=EorzeaLang.JA
        ) == '霧'
        assert EorzeaWeather.forecast(
            'Eureka Pagos',
            1542651599.999,
            lang=EorzeaLang.DE
        ) == 'Neblig'
        assert EorzeaWeather.forecast(
            'Eureka Pagos',
            1542651599.999,
            lang=EorzeaLang.FR
        ) == 'Brouillard'



import re

try:
    from typing import Iterable, List, Literal, overload
    Lang = Literal['en', 'jp', 'de', 'fr']
except:
    from typing import Iterable, List, overload
    Lang = str

from numpy import uint32

from .Data.TerritoryWeather import territory as _territory
from .Data.Weather import weather as _weather
from .Data.WeatherRate import weather_rate as _weather_rate


class EorzeaWeather:
    """
    EoreaWeather
    """

    @overload
    @staticmethod
    def forecast(
        placename: str,
        timestamp: Iterable[float],
        lang: Lang = "en",
        strict: bool = True
    ) -> List[str]:
        """Forecast Eorzea weather by place

        Parameters
        ----------
        placename : str
            a valid Eorzea place name
        timestamp : Iterable[float]
            unix timestamp
        lang : Lang, optional
            Recommend use `EorzeaEnv.EorzeaLang`
            valid lang from ffxiv , by default "en"
        strict : bool, optional
            option of search mode, by default True
            + `True` for strict mode
            + `False` for fuzzy mode

        Returns
        -------
        List[str]
            result of forecast

        Raises
        -------
        ValueError
            when using invalid place name
        """
        ...

    @overload
    @staticmethod
    def forecast(
        placename: str,
        timestamp: float,
        lang: Lang = "en",
        strict: bool = True
    ) -> str:
        """Forecast Eorzea weather by place

        Parameters
        ----------
        placename : str
            a valid Eorzea place name
        timestamp : float
            unix timestamp
        lang : Lang, optional
            Recommend use `EorzeaEnv.EorzeaLang`
            valid lang from ffxiv , by default "en"
        strict : bool, optional
            option of search mode, by default True
            + `True` for strict mode
            + `False` for fuzzy mode

        Returns
        -------
        str
            result of forecast

        Raises
        -------
        ValueError
            when using invalid place name
        """
        ...

    @staticmethod
    def forecast(placename, timestamp, lang="en", strict=True):
        placename = _parse_placename(placename)
        weather_rate = _get_weather_rate(placename, strict)

        if isinstance(timestamp, Iterable):
            target = (_calculate_forecast_target(t) for t in timestamp)
            result = [
                _generate_result(ta, weather_rate, lang) for ta in target
            ]

            return result

        if isinstance(timestamp, float):
            target = _calculate_forecast_target(timestamp)
            result = _generate_result(target, weather_rate, lang)

            return result


def _generate_result(target: int, weather_rate: int, lang: str) -> str:
    for r, w in _weather_rate[weather_rate]:
        if target < r:
            return _weather[w][lang]


def _get_weather_rate(placename: str, strict: bool) -> int:
    weather_rate = None
    try:
        weather_rate = _territory[placename]
    except KeyError:
        if strict:
            raise ValueError('valid Eorzea placename required')

        for p, r in _territory.items():
            if re.search(p, placename):
                weather_rate = r
    finally:
        if weather_rate is None:
            raise ValueError('valid Eorzea placename required')
        return weather_rate


def _parse_placename(placename: str) -> str:
    """
    trim `the`

    Parameters
    ----------
    placename : str
        valid placename from FFXIV starts with `the`

    Returns
    -------
    str
        trimmed placename
    """
    placename = placename.lower()
    check_placename = re.search('^the (.*)', placename)

    if check_placename:
        placename = ''.join(check_placename.groups())

    return placename


def _calculate_forecast_target(lt: float) -> int:
    """
    Thanks to Rogueadyn's SaintCoinach library for this calculation
    --------------
    [1] Do the magic 'cause for calculations

    [2] 16:00 is 0, 00:00 is 8 and 08:00 is 16.

    [3] Take Eorzea days since unix epoch.

    Parameters
    ----------
    lt : float
        local timestamp

    Returns
    -------
    int
        weather period start
    """

    bell = lt / 175
    increment = uint32(bell + 8 - (bell % 8)) % 24
    total_days = uint32(lt / 4200)
    calc_base = total_days * 0x64 + increment
    step1 = uint32(calc_base << 0xB) ^ calc_base
    step2 = (step1 >> 8) ^ step1
    return int(step2 % 100)



territory = {
    'aery': 28,
    'aetherochemical research facility': 0,
    'akadaemia anyder': 0,
    'akh afah amphitheatre': 46,
    'ala mhigo': 0,
    'alexander': 0,
    'amaurot': 105,
    'amdapor keep': 28,
    'amh araeng': 108,
    'anamnesis anyder': 0,
    'antitower': 0,
    'aquapolis': 0,
    'arm of the father': 0,
    'arm of the son': 0,
    'ashfall': 0,
    'astragalos': 0,
    'atlas peak': 23,
    'aurum vale': 0,
    'azim steppe': 85,
    'azys lla': 54,
    "baelsar's wall": 40,
    "bardam's mettle": 0,
    'battlehall': 0,
    'binding coil of bahamut': 0,
    'blessed treasury': 77,
    'blue sky': 0,
    'bokairo inn': 82,
    'bowl of embers': 25,
    'bozjan southern front': 124,
    "brayflox's longstop": 0,
    'breath of the creator': 0,
    'burden of the father': 0,
    'burden of the son': 0,
    'burn': 0,
    'burning heart': 44,
    'carteneau flats: borderland ruins': 0,
    'castrum abania': 0,
    'castrum fluminis': 0,
    'castrum meridianum': 0,
    'central azys lla': 0,
    'central decks': 0,
    'central shroud': 3,
    'central thanalan': 10,
    'chocobo square': 0,
    'chrysalis': 0,
    'churning mists': 52,
    "cid's memory": 123,
    'cinder drift': 28,
    'cloud nine': 0,
    'coerthas central highlands': 21,
    'coerthas western highlands': 49,
    'command room': 14,
    'company workshop - mist': 0,
    'company workshop - shirogane': 0,
    'company workshop - the goblet': 0,
    'company workshop - the lavender beds': 0,
    'confessional of toupasa the elder': 102,
    'containment bay p1t6': 69,
    'containment bay s1t7': 66,
    'containment bay z1t9': 75,
    'copied factory': 0,
    'copperbell mines': 0,
    'core': 0,
    'crown of the immaculate': 104,
    'crystal tower training grounds': 90,
    'crystarium': 112,
    'cuff of the father': 0,
    'cuff of the son': 0,
    "cutter's cry": 0,
    "dalamud's shadow": 0,
    'dancing plague': 103,
    'deltascape v1.0': 88,
    'deltascape v2.0': 88,
    'deltascape v3.0': 88,
    'deltascape v4.0': 88,
    'diadem': 60,
    'dohn mheg': 0,
    'doma castle': 0,
    'doman enclave': 84,
    'dravanian forelands': 50,
    'dravanian hinterlands': 51,
    'drowned city of skalla': 0,
    'dun scaith': 58,
    'dungeons of lyhe ghiah': 0,
    'dusk vigil': 42,
    'dying gasp': 105,
    'dzemael darkhold': 0,
    'east shroud': 4,
    'eastern la noscea': 18,
    'eastern thanalan': 11,
    'eighteenth floor': 0,
    'emanation': 87,
    'empty': 0,
    'endeavor': 14,
    'eorzean alliance headquarters': 0,
    'eorzean subterrane': 0,
    'eulmore': 113,
    'eureka anemos': 91,
    'eureka hydatos': 100,
    'eureka pagos': 94,
    'eureka pyros': 96,
    'excavation tunnels': 0,
    'eyes of the creator': 0,
    "fall of belah'dia": 0,
    'falling city of nym': 0,
    'feasting grounds': 65,
    'fields of glory': 67,
    'firmament': 47,
    'first altar of djanan qhat': 0,
    'fist of the father': 0,
    'fist of the son': 0,
    'flame barracks': 0,
    'fold': 0,
    'fortemps manor': 0,
    'foundation': 47,
    'fractal continuum': 0,
    'fringes': 79,
    "frondale's home for friendless foundlings": 0,
    'gandof thunder plains': 0,
    'gangos': 0,
    'ghimlyt dark': 0,
    'goblet': 33,
    'gold saucer': 0,
    'grand cosmos': 0,
    'great glacier': 46,
    'great gubal library': 0,
    'great hunt': 0,
    'griffin crossing': 0,
    'halatali': 0,
    'hall of summoning': 0,
    'hall of the bestiarii': 0,
    'halo': 0,
    'haukke manor': 0,
    'haunted manor': 0,
    'heart of the creator': 0,
    'heart of the sworn': 0,
    'heaven-on-high': 0,
    "hells' kier": 98,
    "hells' lid": 0,
    "heroes' gauntlet": 125,
    'hidden gorge': 0,
    'holminster switch': 102,
    'holocharts': 0,
    'hourglass': 0,
    'house of the fierce': 84,
    'howling eye': 26,
    'hullbreaker isle': 0,
    'ic-04 main bridge': 0,
    'ic-06 central decks': 0,
    'ic-06 main bridge': 0,
    'ic-06 regeneration grid': 0,
    'idyllshire': 55,
    'il mheg': 109,
    'imperial palace': 0,
    'intercessory': 0,
    'interdimensional rift': 92,
    'jade stoa': 93,
    'keeper of the lake': 74,
    'kholusia': 107,
    'kienkan': 0,
    'kobai goten apartment': 0,
    'kobai goten apartment lobby': 0,
    'kugane': 82,
    'kugane castle': 0,
    'kugane ohashi': 0,
    'labyrinth of the ancients': 0,
    'lakeland': 106,
    'lavender beds': 34,
    "leofard's chambers": 0,
    'lichenweed': 73,
    'lightfeather proving grounds': 0,
    'lily hills apartment': 0,
    'lily hills apartment lobby': 0,
    'liminal space': 0,
    'limitless blue': 28,
    'limsa lominsa': 0,
    'limsa lominsa lower decks': 15,
    'limsa lominsa upper decks': 14,
    'lochs': 81,
    'lost canals of uznair': 0,
    'lost city of amdapor': 40,
    'lotus stand': 0,
    'lower aetheroacoustic exploratory site': 0,
    'lower la noscea': 17,
    'lyhe mheg': 40,
    'maelstrom barracks': 0,
    'main bridge': 0,
    "malikah's well": 102,
    'manderville tables': 0,
    "matoya's cave": 0,
    'middle la noscea': 16,
    'misery': 36,
    'mist': 14,
    'mizzenmast inn': 0,
    'mor dhona': 22,
    'mordion gaol': 0,
    'mt. gulg': 102,
    'nabaath mines': 0,
    'navel': 23,
    'nereus trench': 38,
    'neverreap': 0,
    'new gridania': 1,
    'north shroud': 6,
    'northern thanalan': 13,
    'ocular': 0,
    'old gridania': 2,
    'omega control': 0,
    'onsal hakair': 117,
    'orbonne monastery': 0,
    'outer coil': 28,
    'outer la noscea': 24,
    'palace of the dead': 0,
    'parrock': 0,
    'peaks': 80,
    'pendants personal suite': 112,
    'pharos sirius': 28,
    'pillars': 48,
    'porta decumana': 31,
    'praetorium': 0,
    'prima vista bridge': 0,
    'prima vista tiring room': 0,
    'private chambers - mist': 0,
    'private chambers - shirogane': 0,
    'private chambers - the goblet': 0,
    'private chambers - the lavender beds': 0,
    'private cottage - mist': 0,
    'private cottage - shirogane': 0,
    'private cottage - the goblet': 0,
    'private cottage - the lavender beds': 0,
    'private house - mist': 0,
    'private house - shirogane': 0,
    'private house - the goblet': 0,
    'private house - the lavender beds': 0,
    'private mansion - mist': 0,
    'private mansion - shirogane': 0,
    'private mansion - the goblet': 0,
    'private mansion - the lavender beds': 0,
    'psiscape v1.0': 92,
    'psiscape v2.0': 92,
    "puppets' bunker": 0,
    'qitana ravel': 102,
    'ragnarok': 0,
    'ragnarok central core': 0,
    'ragnarok drive cylinder': 0,
    'ragnarok main bridge': 0,
    "rak'tika greatwood": 110,
    'reisen temple': 0,
    'reisen temple road': 0,
    'resonatorium': 0,
    "rhalgr's reach": 78,
    'rhotano sea': 0,
    'ridorana cataract': 0,
    'ridorana lighthouse': 0,
    'rising stones': 0,
    'roost': 0,
    'royal airship landing': 76,
    'royal city of rabanastre': 0,
    'royal menagerie': 0,
    'royal palace': 0,
    'ruby bazaar offices': 0,
    'ruby sea': 83,
    'ruling chamber': 0,
    'sacrificial chamber': 0,
    "saint endalim's scholasticate": 0,
    "saint mocianne's arboretum": 0,
    'sanctum of the twelve': 0,
    'sastasha': 0,
    'sea of clouds': 53,
    'seal rock': 59,
    'seat of sacrifice': 125,
    'seat of the first bow': 0,
    'seat of the lord commander': 0,
    'shifting altars of uznair': 0,
    'shirogane': 82,
    'shisui of the violet tides': 0,
    'sigmascape v1.0': 92,
    'sigmascape v2.0': 92,
    'sigmascape v3.0': 92,
    'sigmascape v4.0': 92,
    'singularity reactor': 56,
    'sirensong sea': 36,
    'snowcloak': 42,
    'sohm al': 0,
    'sohr khai': 0,
    'soul of the creator': 0,
    'south shroud': 5,
    'southern thanalan': 12,
    'steps of faith': 28,
    'stone vigil': 27,
    'striking tree': 43,
    "sultana's breath apartment": 0,
    "sultana's breath apartment lobby": 0,
    'sunken temple of qarn': 0,
    "swallow's compass": 0,
    'syrcus tower': 0,
    'syrcus trench': 0,
    'tam-tara deepcroft': 0,
    'tempest': 111,
    'temple of the fist': 0,
    'terncliff': 0,
    'terncliff bay': 0,
    'thok ast thok': 57,
    'thornmarch': 30,
    'thousand maws of toto-rak': 0,
    'topmast apartment': 0,
    'topmast apartment lobby': 0,
    'transmission control': 123,
    'transparency': 76,
    "trial's threshold": 0,
    'twin adder barracks': 0,
    'twinning': 0,
    "ul'dah - steps of nald": 7,
    "ul'dah - steps of thal": 8,
    'ultimacy': 26,
    'upper aetheroacoustic exploratory site': 0,
    'upper la noscea': 20,
    'vault': 0,
    'void ark': 37,
    'waking sands': 0,
    "wanderer's palace": 0,
    'weeping city of mhach': 0,
    'weeping saint': 21,
    'western la noscea': 19,
    'western thanalan': 9,
    'whorleater': 38,
    "wolves' den pier": 29,
    'world of darkness': 0,
    'wreath of snakes': 99,
    'xelphatol': 40,
    'yanxia': 84
}
